from typing import List, Tuple

import investor8_sdk
import pandas as pd

from i8_terminal.config import USER_SETTINGS
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_user_watchlists() -> List[Tuple[str, str]]:
    results = investor8_sdk.UserApi().get_watchlists_by_user(user_id=USER_SETTINGS.get("user_id"))
    df = pd.DataFrame([d.to_dict() for d in results.watchlists])[["name"]]
    df["name"] = df["name"].apply(lambda x: f'"{x}"' if " " in x else x)
    df["desc"] = ""
    return list(df.to_records(index=False))


class UserWatchlistsParamType(AutoCompleteChoice):
    name = "userwatchlists"

    def get_suggestions(self, keyword: str, pre_populate: bool = False) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_user_watchlists())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]

        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "USERWATCHLISTS"



from datetime import datetime
from typing import List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice
from i8_terminal.types.fin_period_param_type import FinancialsPeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType


class FinancialsIdentifierParamType(AutoCompleteChoice):
    name = "financials"

    def __init__(self) -> None:
        self._ticker_auto_comp = TickerParamType()
        self._period_auto_comp = FinancialsPeriodParamType()

    def get_suggestions(
        self, keyword: str, pre_populate: bool = False, param_type: str = None
    ) -> List[Tuple[str, str]]:
        if param_type == "ticker":
            return self._ticker_auto_comp.get_suggestions(keyword)
        elif param_type == "year":
            years = [(str(i), "") for i in range(datetime.now().year, 2008, -1)]
            if keyword.strip() == "":
                return years
            return [y for y in years if y[0].startswith(keyword)]
        else:
            return self._period_auto_comp.get_suggestions(keyword, True)

    def __repr__(self) -> str:
        return "FINANCIALS"



from typing import Dict, Optional

import click
import numpy as np
import pandas as pd
from pandas import DataFrame
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from i8_terminal.commands.company import company
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import format_metrics_df
from i8_terminal.common.metrics import get_current_metrics_df
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.config import APP_SETTINGS, get_table_style
from i8_terminal.types.ticker_param_type import TickerParamType


def get_section_stock_infos_df(tickers: str, target: str, section: Dict[str, str]) -> Optional[DataFrame]:
    df = get_current_metrics_df(tickers, section["metrics"])
    if df is None:
        return None
    if section["name"] == "Financials":
        fyq_rows = []
        for ticker, ticker_df in df.groupby("Ticker"):
            fyq_rows.append(
                {
                    "Ticker": ticker,
                    "metric_name": "fyq",
                    "value": ticker_df["period"].values[0],
                    "display_name": "FYQ",
                    "data_format": "str",
                    "display_format": "str",
                }
            )
        df = pd.concat([pd.DataFrame(fyq_rows), df], ignore_index=True, axis=0)
    formatted_df = format_metrics_df(df, target)
    formatted_df = formatted_df.pivot(index="display_name", columns="Ticker", values="value").reset_index(level=0)
    formatted_df["Section"] = section["name"]
    sorter = df["display_name"].unique()
    sorter_index = dict(zip(sorter, range(len(sorter))))
    formatted_df["Rank"] = formatted_df["display_name"].map(sorter_index)
    formatted_df.sort_values("Rank", inplace=True)
    formatted_df.drop("Rank", axis=1, inplace=True)
    return formatted_df


def get_stock_infos_df(tickers: str, target: str) -> Optional[DataFrame]:
    return (
        pd.concat(
            [
                get_section_stock_infos_df(tickers, target, section)
                for section in APP_SETTINGS["commands"]["company_compare"]["metric_groups"]
            ]
        )
        .rename(columns={"display_name": "Name"})
        .astype(object)
        .replace(np.nan, "N/A")
    )


def companies_df2tree(df: DataFrame, tickers: str) -> Tree:
    tickers_list = tickers.replace(" ", "").upper().split(",")
    col_width = 40
    plot_title = f"Comparison of {', '.join(tickers_list)}"
    plot_title = " and ".join(plot_title.rsplit(", ", 1))
    tree = Tree(Panel(plot_title, width=50))
    # Add header table to tree
    header_table = Table(
        width=50 + (col_width * (len(tickers_list) - 1)), show_lines=False, show_header=False, box=None
    )
    header_table.add_column(width=35, style="magenta")
    for p in tickers_list:
        header_table.add_column(width=col_width, justify="right", style="magenta")
    header_table.add_row("Ticker", *tickers_list)
    tree.add(header_table)

    table_style = get_table_style("company_compare")

    for sec_name, sec_values in df.groupby("Section", sort=False):
        sec_branch = tree.add(f"[magenta]{sec_name}")
        for i, r in sec_values.reset_index().iterrows():
            t = Table(
                width=46 + (col_width * (len(tickers_list) - 1)),
                show_lines=False,
                show_header=False,
                box=None,
                row_styles=[table_style["row_styles"][i % 2]],
            )
            t.add_column(width=31)
            for tk in tickers_list:
                t.add_column(width=col_width, justify="right")
            t.add_row(r["Name"], *[f"{d}" for d in r[tickers_list].values])
            sec_branch.add(t)

    return tree


def export_companies_data(
    export_df: pd.DataFrame,
    export_path: str,
) -> None:
    console = Console()
    extension = export_path.split(".")[-1]
    if extension == "csv":
        export_df.drop(columns=["Section"], inplace=True)  # type: ignore
        export_df.to_csv(export_path, index=False)
        console.print(f"Data is saved on: {export_path}")
    elif extension == "xlsx":
        writer = pd.ExcelWriter(export_path, engine="xlsxwriter")
        for sec_name, sec_values in export_df.groupby("Section", sort=False):
            sec_values.drop(columns=["Section"], inplace=True)  # type: ignore
            sec_values.rename(columns={"Name": ""}, inplace=True)
            sec_values.to_excel(writer, sheet_name=sec_name, startrow=1, header=False, index=False)
            workbook = writer.book
            column_width = 18
            header_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["header"])
            metric_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["metric"])
            column_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["company"]["column"])
            worksheet = writer.sheets[sec_name]
            headers = sec_values.columns.tolist()
            for col_num, value in enumerate(headers):
                if type(value) is tuple:
                    worksheet.merge_range(0, col_num, len(value) - 1, col_num, " ".join(reversed(value)), header_format)
                else:
                    worksheet.write(0, col_num, value, header_format)
            worksheet.set_column(0, 0, 20, metric_format)
            worksheet.set_column(1, len(sec_values.columns) - 1, column_width, column_format)
        writer.save()
        console.print(f"Data is saved on: {export_path}")
    else:
        console.print("export_path is not valid")


@company.command()
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def compare(tickers: str, export_path: Optional[str]) -> None:
    """
    Compare details of the given companies.

    Examples:

    `i8 company compare --tickers MSFT,AAPL`

    """
    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        stock_infos_df = get_stock_infos_df(tickers, target="store" if export_path else "console")
        if stock_infos_df is None:
            status.stop()
            click.echo("No data found!")
            return
    if export_path:
        export_companies_data(
            stock_infos_df,
            export_path,
        )
    else:
        tree = companies_df2tree(stock_infos_df, tickers)
        console.print(tree)



from pydoc import locate
from typing import Any, Dict, List

import click
import investor8_sdk
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from pandas import DataFrame
from plotly.subplots import make_subplots
from rich.console import Console

from i8_terminal.app.layout import get_plot_default_layout
from i8_terminal.app.plot_server import serve_plot
from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import get_click_command_path, pass_command
from i8_terminal.common.layout import df2Table, format_metrics_df
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import PlotType
from i8_terminal.types.chart_param_type import ChartParamType
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.output_param_type import OutputParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def get_historical_metrics_df(
    tickers: List[str],
    metrics: List[str],
) -> DataFrame:
    historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
        symbols=",".join(tickers), metrics=",".join(metrics), from_period_offset=-10, to_period_offset=0
    )
    df = pd.DataFrame.from_records(
        [
            (ticker, metric, period_value.period, period_value.period_date_time, period_value.value)
            for ticker, metric_dict in historical_metrics.data.items()
            for metric, period_value_list in metric_dict.items()
            for period_value in period_value_list
        ],
        columns=["Ticker", "metric_name", "Period", "PeriodDateTime", "Value"],
    )
    metadata_df = pd.DataFrame([h.to_dict() for h in historical_metrics.metadata])
    df = pd.merge(df, metadata_df, on="metric_name")
    df.rename(columns={"display_name": "Metric", "Value": "value"}, inplace=True)
    df["value"] = df.apply(
        lambda metric: locate(metric.data_format)(locate("float")(metric.value) if metric.data_format == "int" else metric.value), axis=1  # type: ignore
    )
    return df


def create_fig(
    df: DataFrame,
    cmd_context: Dict[str, Any],
    tickers: List[str],
    chart_type: str,
) -> go.Figure:
    vertical_spacing = 0.02
    layout = dict(
        title=cmd_context["plot_title"],
        autosize=True,
        hovermode="closest",
        legend=dict(font=dict(size=11), orientation="v"),
        margin=dict(b=20, l=50, r=65),
    )
    metrics = list(set(df["Metric"]))
    rows_num = len(metrics)

    if rows_num == 2:
        row_width = [0.5, 0.5]
    else:
        row_width = [1]

    fig = make_subplots(
        rows=rows_num,
        cols=1,
        shared_xaxes=False if len(list(set(df["default_period_type"]))) > 1 else True,
        vertical_spacing=vertical_spacing,
        row_width=row_width,
    )

    for metric, metric_df in df.groupby("Metric", sort=False):
        metric_idx = metrics.index(metric)
        for ticker, ticker_df in metric_df.sort_values(["PeriodDateTime"]).groupby("Ticker"):
            idx = tickers.index(ticker)
            if chart_type == "bar":
                fig.add_trace(
                    go.Bar(
                        x=ticker_df["Period"],
                        y=ticker_df["value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
            else:
                fig.add_trace(
                    go.Scatter(
                        x=ticker_df["Period"],
                        y=ticker_df["value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
        if len(metrics) > 1:
            fig["layout"][f"xaxis{metric_idx+1}"]["dtick"] = round(len(metric_df["Period"]) / 10)

    fig.update_traces(hovertemplate="%{y} %{x}")
    df["Reversed Period"] = df.apply(lambda row: f"{row.Period.split(' ')[1]} {row.Period.split(' ')[0]}", axis=1)
    sorted_periods = [
        f"{reversed_period.split(' ')[1]} {reversed_period.split(' ')[0]}"
        for reversed_period in sorted(set(df["Reversed Period"]))
    ]
    fig.update_xaxes(
        rangeslider_visible=False,
        spikemode="across",
        spikesnap="cursor",
        categoryorder="array",
        categoryarray=sorted_periods,
    )

    fig.update_layout(
        **layout,
        **get_plot_default_layout(),
        legend_title_text=None,
        xaxis_title=None,
        yaxis_title=None,
    )

    # Add yaxis titles
    for idx, r in enumerate(np.cumsum(row_width)[::-1]):
        fig["layout"][f"yaxis{idx+1}"]["title"] = metrics[idx]

    fig.update_annotations(
        dict(
            font_size=10,
            font_color="#525252",
        )
    )
    fig.update_yaxes(
        title_font_size=10,
    )

    return fig


@metrics.command()
@click.pass_context
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option(
    "--metrics",
    "-m",
    type=MetricParamType(),
    default="pe_ratio_ttm",
    help="Comma-separated list of daily metrics.",
)
@click.option(
    "--output",
    "-o",
    type=OutputParamType(),
    default="terminal",
    help="Output can be terminal or plot type.",
)
@click.option(
    "--plot_type",
    "-p",
    type=ChartParamType([("bar", "Bar chart"), ("line", "Line chart")]),
    default="line",
    help="Plot can be bar or line chart.",
)
@pass_command
def historical(
    ctx: click.Context,
    tickers: str,
    metrics: str,
    output: str,
    plot_type: str,
) -> None:
    """
    Compares and plots daily metrics of given companies. TICKERS is a comma-separated list of tickers.

    Examples:

    `i8 metrics historical --metrics price_to_earnings --period 5Y --tickers AMD,INTC,QCOM`
    """
    metrics_list = metrics.replace(" ", "").split(",")
    if not output in ["terminal", "plot"]:
        click.echo(click.style(f"`{output}` is not valid output type.", fg="yellow"))
        return
    if output == "plot" and len(metrics_list) > 2:
        click.echo(click.style("For the `plot` output type you can enter up to only two metrics.", fg="yellow"))
        return
    command_path_parsed_options_dict = {
        "--tickers": tickers,
        "--metrics": metrics,
        "--output": output,
        "--plot_type": plot_type,
    }
    command_path = get_click_command_path(ctx, command_path_parsed_options_dict)
    tickers_list = tickers.replace(" ", "").upper().split(",")
    if len(tickers_list) > 5:
        click.echo(click.style("You can enter up to 5 tickers.", fg="yellow"))
        return
    if not plot_type in ["bar", "line"]:
        click.echo(click.style(f"`{plot_type}` is not valid chart type.", fg="yellow"))
        return
    cmd_context = {
        "command_path": command_path,
        "tickers": tickers_list,
        "plot_type": PlotType.CHART.value,
    }

    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        df = get_historical_metrics_df(tickers_list, metrics_list)
        if output == "plot":
            cmd_context["plot_title"] = f"Historical {' and '.join(list(set(df['Metric'])))}"
            status.update("Generating plot...")
            fig = create_fig(df, cmd_context, tickers_list, plot_type)

    if output == "plot":
        serve_plot(fig, cmd_context)
        return

    columns_justify: Dict[str, Any] = {}
    for metric_display_name, metric_df in df.groupby("Metric"):
        columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
    formatted_df = (
        format_metrics_df(df, "console")
        .pivot(index=["Ticker", "Period"], columns="Metric", values="value")
        .reset_index()
    )
    formatted_df["reversed_period"] = formatted_df.apply(
        lambda row: f"{row.Period.split(' ')[1]} {row.Period.split(' ')[0]}", axis=1
    )
    formatted_df.sort_values(["Ticker", "reversed_period"], ascending=False, inplace=True)
    formatted_df.drop(columns=["reversed_period"], inplace=True)
    table = df2Table(
        formatted_df,
        columns_justify=columns_justify,
    )
    console.print(table)



import logging
import os
import signal
import sys
from threading import Timer
from typing import Any

import click
import investor8_sdk
from flask import Flask, redirect, request

from i8_terminal.config import APP_SETTINGS, save_user_settings

app = Flask(__name__)


def _configure_flask() -> None:
    # TODO: the following method does not disable dash logging (it should)
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)
    log.disabled = True

    cli = sys.modules["flask.cli"]
    cli.show_server_banner = lambda *x: None  # type: ignore


@app.route("/")
def login() -> Any:
    verification_code = request.args.get("verificationCode")
    if verification_code:
        body = {"ReqId": app.config["REQUEST_ID"], "VerificationCode": verification_code}
        try:
            resp = investor8_sdk.UserApi().login_with_code(body=body)
            user_setting = {
                "i8_core_token": resp.token,
                "i8_core_api_key": resp.api_key,
                "user_id": resp.user_id,
                "allow_terminal_logging": resp.allow_terminal_logging,
            }
            save_user_settings(user_setting)
            click.echo("User logged in successfully!")
            Timer(4.0, shutdown_server).start()
            return redirect("https://www.investoreight.com/account/loginsuccessful")
        except Exception:
            click.echo("Login failed. Invalid request-id or verification code.")
    else:
        click.echo("Missing the required parameter 'verificationCode'!")
        shutdown_server()


def shutdown_server() -> None:
    os.kill(os.getpid(), signal.SIGINT)  # Shutdown server


def run_server(req_id: str) -> None:
    _configure_flask()
    app.config["REQUEST_ID"] = req_id
    app.debug = False
    app.run(host="localhost", port=APP_SETTINGS["app"]["port"])



import os
from typing import Dict, List, Optional

import arrow
import investor8_sdk
import numpy as np
import pandas as pd
from investor8_sdk import MetricsApi
from pandas import DataFrame, read_csv

from i8_terminal.common.layout import format_metrics_df
from i8_terminal.common.utils import is_cached_file_expired, similarity
from i8_terminal.config import APP_SETTINGS, SETTINGS_FOLDER


def get_indicators_list(indicator: Optional[str] = None) -> List[str]:
    indicators_dict = {
        "Momentum": ["ma5", "ma12", "ma26", "ma52", "ema5", "ema12", "ema26", "ema52"],
        "RSI": ["rsi_7d", "rsi_14d", "rsi_1m", "rsi_3m"],
        "Alpha": ["alpha_1w", "alpha_2w", "alpha_1m", "alpha_3m", "alpha_6m", "alpha_1y", "alpha_2y", "alpha_5y"],
        "Beta": ["beta_1w", "beta_2w", "beta_1m", "beta_3m", "beta_6m", "beta_1y", "beta_2y", "beta_5y"],
        "Volume": ["volume"],
    }
    if indicator:
        return indicators_dict.get(indicator, [""])
    else:
        return [item for sublist in indicators_dict.values() for item in sublist]


def find_similar_fin_metric(metric: str) -> Optional[str]:
    metrics_meta_data = get_all_metrics_df()[["metric_name"]]
    best_match_similarity = 0.0
    best_match = ""
    for idx, m in metrics_meta_data.iterrows():
        sim = similarity(metric.upper(), m["metric_name"].upper())
        if sim > best_match_similarity:
            best_match_similarity = sim
            best_match = m["metric_name"]
    if best_match_similarity < APP_SETTINGS["metrics"]["similarity_threshold"]:
        return None
    return best_match


def find_similar_indicator(indicator: str) -> Optional[str]:
    default_indicators = {"rsi": "rsi_14d", "alpha": "alpha_1y", "beta": "beta_1y"}
    defualt_ind = default_indicators.get(indicator)
    if defualt_ind:
        return defualt_ind
    indicators_list = get_indicators_list()
    best_match_similarity = 0.0
    best_match = ""
    for ind in indicators_list:
        sim = similarity(indicator, ind)
        if sim > best_match_similarity:
            best_match_similarity = sim
            best_match = ind
    if best_match_similarity < APP_SETTINGS["metrics"]["similarity_threshold"]:
        return None

    return best_match


def get_all_metrics_df() -> DataFrame:
    metric_path = f"{SETTINGS_FOLDER}/metrics_metadata.csv"
    if os.path.exists(metric_path) and not is_cached_file_expired(metric_path):
        df = read_csv(metric_path)
    else:
        all_metrics = MetricsApi().get_list_metrics_metadata(page_size=1000)
        df = DataFrame([m.to_dict() for m in all_metrics])
        df["categories"] = [str(cat) for cat in df["categories"]]
        df = df.drop(columns=["id", "last_modified"])
        df.to_csv(metric_path, index=False)

    return df


def get_all_financial_metrics_df() -> DataFrame:
    metric_path = f"{SETTINGS_FOLDER}/financial_metrics_metadata.csv"
    if os.path.exists(metric_path) and not is_cached_file_expired(metric_path):
        df = read_csv(metric_path)
    else:
        all_metrics = MetricsApi().get_list_financial_metrics_metadata()
        df = DataFrame([m.to_dict() for m in all_metrics])
        df.to_csv(metric_path, index=False)

    return df


def get_metrics_display_names(metrics: List[str]) -> List[str]:
    all_metrics = get_all_metrics_df()[["metric_name", "display_name"]]
    return list(set(all_metrics[all_metrics.metric_name.isin(metrics)]["display_name"]))


def get_metric_info(name: str) -> Dict[str, str]:
    all_metrics = get_all_metrics_df()
    metric = all_metrics.loc[all_metrics.metric_name == name].replace(np.nan, "", regex=True).to_dict("records")[0]
    return {
        "display_name": metric["display_name"],
        "unit": metric["unit"],
        "type": metric.get("type", ""),
        "display_format": metric.get("display_format", ""),
        "default_period_type": metric.get("period_type_default", ""),
        "description": metric.get("description", "No Description"),
        "remarks": metric.get("remarks"),
    }


def get_period_start_date(period: str) -> str:
    period_start_date: Dict[str, str] = {
        "1M": arrow.now().shift(months=-1).datetime.strftime("%Y-%m-%d"),
        "3M": arrow.now().shift(months=-3).datetime.strftime("%Y-%m-%d"),
        "6M": arrow.now().shift(months=-6).datetime.strftime("%Y-%m-%d"),
        "1Y": arrow.now().shift(years=-1).datetime.strftime("%Y-%m-%d"),
        "3Y": arrow.now().shift(years=-3).datetime.strftime("%Y-%m-%d"),
        "5Y": arrow.now().shift(years=-5).datetime.strftime("%Y-%m-%d"),
    }
    return period_start_date[period]


def get_current_metrics_df(tickers: str, metricsList: str) -> Optional[pd.DataFrame]:
    metrics = investor8_sdk.MetricsApi().get_current_metrics(
        symbols=tickers,
        metrics=metricsList,
    )
    if metrics.data is None:
        return None
    metrics_data_df = pd.DataFrame([m.to_dict() for m in metrics.data])
    metrics_data_df.rename(columns={"metric": "metric_name", "symbol": "Ticker"}, inplace=True)
    metrics_metadata_df = pd.DataFrame([m.to_dict() for m in metrics.metadata])
    return pd.merge(metrics_data_df, metrics_metadata_df, on="metric_name").replace("string", "str")


def prepare_current_metrics_formatted_df(df: DataFrame, target: str) -> DataFrame:
    formatted_df = format_metrics_df(df, target)
    return (
        formatted_df.pivot(index="Ticker", columns="display_name", values="value")
        .reset_index(level=0)
        .reindex(np.insert(df["display_name"].unique(), 0, "Ticker"), axis=1)
    )



from typing import List, Tuple

import investor8_sdk

from i8_terminal.common.stock_info import get_stocks
from i8_terminal.config import USER_SETTINGS
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_user_watchlist_tickers() -> List[Tuple[str, str]]:
    results = investor8_sdk.UserApi().get_watchlists_by_user(user_id=USER_SETTINGS.get("user_id"))
    tickers = set(ticker for wl in results.watchlists for ticker in wl.tickers)
    stocks = get_stocks()
    return [(tk, name) for (tk, name) in stocks if tk in tickers]


class UserWatchlistTickersParamType(AutoCompleteChoice):
    name = "userwatchlisttickers"

    def get_suggestions(self, keyword: str, pre_populate: bool = False) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_user_watchlist_tickers())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]

        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "USERWATCHLISTTICKERS"
